#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include <string.h>

#include "pce.h"
#include "mstring.h"
#include "mvector.h"
#include "types.h"
#include "forms.h"


static inline int min(int a, int b)
{
	return a <= b ? a : b;
}

static inline int max(int a, int b)
{
	return a >= b ? a : b;
}

namespace forms
{
	int form_list(mstring title, mvector<mstring>& entries)
	{
		const int free_space = 240 - 22 - 15;
		const int max_items = free_space / 15;
		const int y0 = 22 + (free_space - max_items * 15) / 2;

		int first = 0, count = entries.size(), focus = 0;

		using namespace gp2x::joystick;

		for (;;)
		{
			if (focus < first) first = focus;
			if (focus >= (first + max_items)) first = max(0, focus - max_items + 1);

			mstring status;
			status << (count > 0 ? focus+1 : 0) << " of " << count << " - Press Y to exit";

				surf->rect(0, 0, 320, 240, 0);

				surf->print(font_7x13b, (320 - surf->text_width(font_7x13b, title, false)) / 2, 4, title, 255, 255, false);

				for (int i = first, last = first + min(max_items, count - first), y = y0; i < last; i++)
				{
					if (focus == i) surf->rect(2, y, 320 - 4, 14, surf->color(128, 128, 128));
					surf->print(font_7x13, 4, y, entries[i], 255, 255, false);

					y += 15;
				}

				{
					const int tw = surf->text_width(font_7x13, status, false);
					const int x = (320 - tw) / 2;

					surf->print(font_7x13, x, 240 - 15, status, 255, 255, false);
				}

				surf->flip();

				for (bool redraw = false; !redraw; )
				{
					int joy = gp2x::joystick::wait_button_press();

					if (joy & BUTTON_Y)
					{
						return -1;
					}
					else if ((joy & BUTTON_DOWN) && focus < count-1)
					{
						focus++;
						redraw = true;
					}
					else if ((joy & BUTTON_UP) && focus > 0)
					{
						focus--;
						redraw = true;
					}
					else if ((joy & BUTTON_R) && focus < count-1)
					{
						focus = min(focus + max_items, count - 1);
						redraw = true;
					}
					else if ((joy & BUTTON_L) && focus > 0)
					{
						focus = max(focus - max_items, 0);
						redraw = true;
					}
					else if ((joy & BUTTON_A) && count > 0)
					{
						return focus;
					}
				}
		}

		return -1;
	}

	mstring compose_path(const mstring& folder, const mstring& filename)
	{
		if (folder.empty()) return filename;
		if (folder[folder.size() - 1] == '/') return folder + filename;
		return (folder + "/") + filename;
	}

	bool read_files(const mstring& folder, mvector<mstring>& files)
	{
		DIR *dir = opendir(folder);
		if (!dir) return false;

		struct dirent *file;
		while ((file = readdir(dir)) != 0)
		{
			struct stat stat_buf;

			int len = strlen(file->d_name);
			if (len < 5 || strcmp(file->d_name + len - 4, ".pce") && strcmp(file->d_name + len - 4, ".PCE")) continue;

			if (stat(compose_path(folder, mstring(file->d_name)), &stat_buf)) continue;

			if (!S_ISREG(stat_buf.st_mode)) continue;

			files.append(mstring(file->d_name));
		}

		closedir(dir);
		return true;
	}

	mstring DialogFileOpen(const mstring& folder)
	{
		mvector<mstring> files;
		if (!read_files(folder, files)) return "";
		int selected = form_list("Select a ROM", files);
		if (selected < 0) return "";
		return compose_path(folder, files[selected]);
	}
};
