#ifndef GP2X_H
#define GP2X_H

namespace gp2x
{
	namespace joystick
	{
		enum
		{
			BUTTON_UP		= 1,		BUTTON_LEFT		= 1 << 2,	BUTTON_DOWN = 1 << 4,	BUTTON_RIGHT	= 1 << 6,
			BUTTON_START	= 1 << 8,	BUTTON_SELECT	= 1 << 9,	BUTTON_L	= 1 << 10,	BUTTON_R		= 1 << 11,
			BUTTON_A		= 1 << 12,	BUTTON_B		= 1 << 13,	BUTTON_X	= 1 << 14,	BUTTON_Y		= 1 << 15,
			BUTTON_VOL_DOWN	= 1 << 22,	BUTTON_VOL_UP	= 1 << 23,	BUTTON_PUSH	= 1 << 27,
			BUTTON_ALL		= 0x08c0ff55
		};

		int read_state();
		int wait_button_press();
	}

	namespace timer
	{
		enum { HZ = 7372800 };

		void pause(unsigned int ticks);
		unsigned int read_ticks();
	}

	namespace sound
	{
		typedef void (*sound_producer_func)(unsigned short *buffer, int samples);

		void setup(int sound_freq, bool stereo, sound_producer_func producer);
		void play();
		void pause();
		void volume(int left, int right); // 0..100
		void lock();
		void unlock();
	}

	namespace video
	{
		int sync_fps(unsigned fps);
		void wait_v_sync();
		void wait_h_sync();

		class surface8
		{
			unsigned char *m_virtual, *m_physical, *m_data;
			int m_size, m_size1, m_width, m_height;
			bool m_double_buffered, m_first_buffer;

		public:

			surface8(int width, int height, bool hardware, bool double_buffered); // double_buffered only for hardware
			~surface8();

			void flip();
			void flip(int x, int y);
			void flip(int x, int y, int w, int h); // scale to w x h

			void blit(surface8 *source, int sx, int sy, int sw, int sh, int dx, int dy);

			void rect(int x, int y, int w, int h, unsigned char color);
			int text_width(const unsigned char *font, const char *text, bool fixed_width);
			int print(const unsigned char *font, int x, int y, const char *text, unsigned char color_fg, unsigned char color_bg, bool fixed_width);
			void clear(bool both);

			unsigned char color(int r, int g, int b);
			unsigned char *data() { return m_data; }
			int height() const { return m_height; }
			int width() const { return m_width; }
			int pitch() const { return m_width; }
		};
	}
};



#endif
