#include <string.h>
#include "types.h"
#include "cpu.h"
#include "video.h"
#include "sound.h"
#include "io.h"

#define	VRR	2

#define	NODATA	0xff
#define	ENABLE	1
#define	DISABLE	0

const byte TIRQ = 4;

/*
#define	INT_TIMER 4
#define	INT_IRQ2	8
*/

IO io;

extern int snd_use_sound;

int JoyStick(byte *JS);


static byte IO_read_0000()
{
	byte ret = io.vdc_status;
	io.vdc_status=0;
	return ret;
}

static byte IO_read_0001()
{
	return 0;
}

static byte IO_read_0002()
{
	if (io.vdc_reg==VRR) return VRAM[io.VDC[MARR].W*2];
	return io.VDC[io.vdc_reg].B.l;
}

static byte IO_read_0003()
{
	if (io.vdc_reg==VRR) {
		byte ret = VRAM[io.VDC[MARR].W*2+1];
		io.VDC[MARR].W+=io.vdc_inc;
		return ret;
	}
	return io.VDC[io.vdc_reg].B.h;
}

static byte IO_read_0404()
{
	return io.VCE[io.vce_reg.W].B.l;
}

static byte IO_read_0405()
{
	return io.VCE[io.vce_reg.W++].B.h;
}

static byte IO_read_0c00()
{
	return io.timer_counter;
}

static byte IO_read_1000()
{
	byte ret = io.JOY[io.joy_counter]^0xff;
	if (io.joy_select&1) ret>>=4;
	else { ret&=15; io.joy_counter=(io.joy_counter+1)%5; }
	return ret|0; /* country 0:JPN 1=US */
}

static byte IO_read_1402()
{
	return irq_mask;
}

static byte IO_read_1403()
{
	byte ret = irq_status;
	irq_status=0;
	return ret;
}

static void IO_write_1402(byte v)
{
	irq_mask = v;
}

static void IO_write_1403(byte v)
{
	irq_status = (irq_status & ~TIRQ) | (v & 0xf8);
}

static byte IO_read_1803()
{
	//return io.backup = DISABLE;
	return DISABLE;
}

static void IO_write_0000(byte V)
{
	io.vdc_reg = V&31;
}

static void IO_write_0002(byte V)
{
	//countw2[io.vdc_reg]++;
	switch(io.vdc_reg)
	{
	case VWR:
		io.vdc_ratch = V;
		return;
	case HDR:
		io.screen_w = (V+1)*8;
		break;
	case MWR:
		{
			static byte bgw[]={32,64,128,128};
			io.bg_h=(V&0x40)?64:32;
			io.bg_w=bgw[(V>>4)&3];
		}
		break;
	case BYR:
		if (!scroll)
		{
			oldScrollX = ScrollX;
			oldScrollY = ScrollY;
			oldScrollYDiff = ScrollYDiff;
		}
		io.VDC[BYR].B.l = V;
		scroll=1;
		ScrollYDiff=scanline-1;
		return;
	case BXR:
		if (!scroll)
		{
			oldScrollX = ScrollX;
			oldScrollY = ScrollY;
			oldScrollYDiff = ScrollYDiff;
		}
		io.VDC[BXR].B.l = V;
		scroll=1;
		return;
	}
	io.VDC[io.vdc_reg].B.l = V;
	return;
}

static void IO_write_0003(byte V)
{
	switch(io.vdc_reg)
	{
	case VWR:
		VRAM[io.VDC[MAWR].W*2]=io.vdc_ratch;
		VRAM[io.VDC[MAWR].W*2+1]=V;
		vchange[io.VDC[MAWR].W/16]=1;
		vchanges[io.VDC[MAWR].W/64]=1;
		io.VDC[MAWR].W+=io.vdc_inc;
		io.vdc_ratch=0;
		return;
	case VDW:
		io.VDC[VDW].B.h = V;
		io.screen_h = (io.VDC[VDW].W&511)+1;
		MaxLine = io.screen_h-1;
		return;
	case LENR:
		io.VDC[LENR].B.h = V;
		memcpy(VRAM+io.VDC[DISTR].W*2,VRAM+io.VDC[SOUR].W*2,
			(io.VDC[LENR].W+1)*2);
		memset(vchange+io.VDC[DISTR].W/16,1,(io.VDC[LENR].W+1)/16);
		memset(vchange+io.VDC[DISTR].W/64,1,(io.VDC[LENR].W+1)/64);
		io.VDC[DISTR].W += io.VDC[LENR].W+1;
		io.VDC[SOUR].W += io.VDC[LENR].W+1;
		io.vdc_status|=VDC_DMAfinish;
		return;
	case CR :
		{static byte incsize[]={1,32,64,128};
		io.vdc_inc = incsize[(V>>3)&3];
		}
		break;
	case HDR:
		break;
	case BYR:
		if (!scroll) {
			oldScrollX = ScrollX;
			oldScrollY = ScrollY;
			oldScrollYDiff = ScrollYDiff;
		}
		io.VDC[BYR].B.h = V&1;
		scroll=1;
		{
			ScrollYDiff=scanline-1;
		}
		return;
	case SATB:
		io.VDC[SATB].B.h = V;
		io.vdc_satb=1;
		io.vdc_status&=~VDC_SATBfinish;
		return;
	case BXR:
		if (!scroll) {
			oldScrollX = ScrollX;
			oldScrollY = ScrollY;
			oldScrollYDiff = ScrollYDiff;
		}
		io.VDC[BXR].B.h = V & 3;
		scroll=1;
		return;

	}
	io.VDC[io.vdc_reg].B.h = V;
	return;
}

static void IO_write_0400(byte V)
{
	io.vce_cr = V;
}

static void IO_write_0402(byte V)
{
	io.vce_reg.B.l = V;
}

static void IO_write_0403(byte V)
{
	io.vce_reg.B.h = V&1;
}

static void IO_write_0404(byte V)
{
	io.VCE[io.vce_reg.W].B.l= V;
	{
		byte c; int i,n;
		n = io.vce_reg.W;
		c = io.VCE[n].W>>1;
		if (n==0) {
			for(i=0;i<256;i+=16) Pal[i]=c;
		}
		else if (n&15) Pal[n] = c;
	}
}

static void IO_write_0405(byte V)
{
	io.VCE[io.vce_reg.W].B.h = V;
	{
		byte c; int i,n;
		n = io.vce_reg.W;
		c = io.VCE[n].W>>1;
		if (n==0) {
			for(i=0;i<256;i+=16) Pal[i]=c;
		}
		else if (n&15) Pal[n] = c;
	}
	io.vce_reg.W=(io.vce_reg.W+1)&0x1FF;
}

/*
static void IO_write_08xx()
{
}
*/

static void IO_write_0c00(byte V)
{
	io.timer_reload = V&127;
}

static void IO_write_0c01(byte V)
{
	V&=1;
	if (V && !io.timer_start)
		io.timer_counter = io.timer_reload;
	io.timer_start = V;
}

static void IO_write_1000(byte V)
{
	io.joy_select = V&1;
	if (V&2) io.joy_counter = 0;
}

static void IO_write_1807(byte)
{
	//io.backup = ENABLE;
}

void timer_tick()
{
	if (io.timer_start)
	{
		if (--io.timer_counter > 128)
		{
			io.timer_counter = io.timer_reload;
			if (!(irq_mask&TIRQ)) irq_status |= TIRQ;
		}
	}
}


void reset_io()
{
	memset(&io, 0, sizeof(IO));
	scanline = 0;
	prevline = 0;
	io.vdc_status=0;
	io.vdc_inc = 1;
	io.minline = 0;
	io.maxline = 255;

	//io.irq_mask = 0;
	//io.psg_volume = 0;
	//io.psg_ch = 0;
	//for (i = 0; i < 6; i++) io.PSG[i][4] = 0x80;
	//CycleOld = 0;
	//Reset6502();
}


void init_io()
{
	set_io_read(0x0000, IO_read_0000);
	set_io_read(0x0001, IO_read_0001);
	set_io_read(0x0002, IO_read_0002);
	set_io_read(0x0003, IO_read_0003);

	set_io_write(0x0000, IO_write_0000);
	set_io_write(0x0002, IO_write_0002);
	set_io_write(0x0003, IO_write_0003);

	set_io_read(0x0404, IO_read_0404);
	set_io_read(0x0405, IO_read_0405);

	set_io_write(0x0400, IO_write_0400);
	set_io_write(0x0402, IO_write_0402);
	set_io_write(0x0403, IO_write_0403);
	set_io_write(0x0404, IO_write_0404);
	set_io_write(0x0405, IO_write_0405);

	set_io_read(0x0800, IO_read_0800);
	set_io_read(0x0801, IO_read_0801);
	set_io_read(0x0802, IO_read_0802);
	set_io_read(0x0803, IO_read_0803);
	set_io_read(0x0804, IO_read_0804);
	set_io_read(0x0805, IO_read_0805);
	set_io_read(0x0806, IO_read_0806);
	set_io_read(0x0807, IO_read_0807);
	set_io_read(0x0808, IO_read_0808);
	set_io_read(0x0809, IO_read_0809);

	set_io_write(0x0800, IO_write_0800);
	set_io_write(0x0801, IO_write_0801);
	set_io_write(0x0802, IO_write_0802);
	set_io_write(0x0803, IO_write_0803);
	set_io_write(0x0804, IO_write_0804);
	set_io_write(0x0805, IO_write_0805);
	set_io_write(0x0806, IO_write_0806);
	set_io_write(0x0807, IO_write_0807);
	set_io_write(0x0808, IO_write_0808);
	set_io_write(0x0809, IO_write_0809);

	set_io_read(0x0c00, IO_read_0c00);

	set_io_write(0x0c00, IO_write_0c00);
	set_io_write(0x0c01, IO_write_0c01);

	set_io_read(0x1000, IO_read_1000);
	set_io_write(0x1000, IO_write_1000);

	set_io_read(0x1402, IO_read_1402);
	set_io_read(0x1403, IO_read_1403);

	set_io_write(0x1402, IO_write_1402);
	set_io_write(0x1403, IO_write_1403);

	set_io_read(0x1803, IO_read_1803);
	set_io_write(0x1807, IO_write_1807);

	reset_io();
}
