#include <string.h>
#include <malloc.h>

namespace memory {

enum { MIN_SIZE = 3, MAX_SIZE = 31 };

struct bucket
{
	bucket *next;
	int size, count;
	char data[1];
};

static bool inited = false;
static bucket *buckets[MAX_SIZE + 1];

static void init()
{
	for (int i = MIN_SIZE; i <= MAX_SIZE; i++) buckets[i] = 0;
}

inline static bucket *ptr_to_bucket(void *ptr)
{
	static bucket bs;
	char *p = (char *) ptr;
	p -= ((char *) &bs.data) - ((char *) &bs);
	return (bucket *) p;
}

char *mem_allocate(int size, int *real_size)
{
	if (!inited) init();
	int size_log = MIN_SIZE;
	while ((1 << size_log) < size) size_log++;
	bucket *ret;
	if (!buckets[size_log])
	{
		ret = (bucket *) malloc(sizeof(bucket) - 1 + (1 << size_log)); 
		ret->size = size_log;
	}
	else 
	{
		ret = buckets[size_log];
		buckets[size_log] = ret->next;
	}
	ret->count = 1;
	if (real_size) *real_size = 1 << size_log;
	return (char *) &(ret->data);
}

void mem_free(void *ptr)
{
	bucket *b = ptr_to_bucket(ptr);
	b->count--;
	if (b->count <= 0)
	{
		b->next = buckets[b->size];
		buckets[b->size] = b;
	}
}

void mem_add_owner(void *ptr)
{
	bucket *b = ptr_to_bucket(ptr);
	b->count++;
}

int mem_get_owner_count(void *ptr)
{
	bucket *b = ptr_to_bucket(ptr);
	return b->count;
}

}
