#include <string.h>
#include <malloc.h>
#include "mstring.h"
#include "memory.h"

static inline int min(int a, int b)
{
	return a <= b ? a : b;
}

static inline int max(int a, int b)
{
	return a >= b ? a : b;
}

void mstring::fork(int n)
{
	if (n > this->capacity || memory::mem_get_owner_count(this->buffer) > 1)
	{
		char *p = this->buffer;
		this->buffer = memory::mem_allocate(max(this->capacity, n), &this->capacity);
		strcpy(this->buffer, p);
		memory::mem_free(p);
	}
}

mstring::mstring(const char *str, int capacity )
{
	if (str == 0)
	{
		this->length = 0;
		this->buffer = memory::mem_allocate(max(1, capacity), &this->capacity);
		this->buffer[0] = 0;
	}
	else
	{
		this->length = (int) strlen(str);
		this->buffer = memory::mem_allocate(max(capacity, this->length + 1), &this->capacity);
		strcpy(this->buffer, str);
	}
}

mstring::mstring(int capacity)
{
	this->length = 0;
	this->buffer = memory::mem_allocate(max(1, capacity), &this->capacity);
	this->buffer[0] = 0;
}

mstring::mstring(const mstring &str, int capacity)
{
	this->length = str.length;
	this->capacity = str.capacity;
	this->buffer = str.buffer;
	memory::mem_add_owner(str.buffer);
}

mstring::mstring(char c, int count)
{
	this->length = count;
	this->buffer = memory::mem_allocate(count + 1, &this->capacity);
	for (int i = 0; i < count; i++) this->buffer[i] = c;
	this->buffer[count] = 0;
}

mstring::~mstring()
{
	memory::mem_free(this->buffer);
}

bool mstring::operator==(const mstring &str) const
{
	return 0 == strcmp(this->buffer, str.buffer);
}

bool mstring::operator!=(const mstring &str) const
{
	return 0 != strcmp(this->buffer, str.buffer);
}

bool mstring::operator==(const char *str) const
{
	return 0 == strcmp(this->buffer, str);
}

bool mstring::operator!=(const char *str) const
{
	return 0 != strcmp(this->buffer, str);
}

mstring mstring::trim() const
{
	char *p1 = this->buffer;
	char *p2 = p1 + this->length - 1;
	
	while (*p1 <= 32 && *p1 > 0) p1++;
	while (p2 >= p1 && *p2 <= 32 && *p2 > 0) p2--;
	
	int ret_size = int(p2 - p1) + 1;
	int ret_capacity;
	
	char *ret_buffer = memory::mem_allocate(ret_size + 1, &ret_capacity);
	if (ret_size > 0) strncpy(ret_buffer, p1, ret_size);
	ret_buffer[ret_size] = 0;

	return mstring(ret_buffer, ret_size, ret_capacity);
}

mstring mstring::substr(int start) const
{
	return substr(start, this->length - start);
}

mstring mstring::substr(int start, int count) const
{
	if (start < 0) start = max(0, this->length + start); else start = min(start, this->length);
	count = (count < 0) ? max(this->length - start + count, 0) : min(count, this->length - start);
	int ret_size = count;
	int ret_capacity;
	
	char *ret_buffer = memory::mem_allocate(ret_size + 1, &ret_capacity);
	if (count > 0) strncpy(ret_buffer, this->buffer + start, count); 
	ret_buffer[count] = 0;
	
	return mstring(ret_buffer, ret_size, ret_capacity);
}

int mstring::find(char c, int start) const
{
	if (start < 0) start = max(0, this->length + start);
	start = min(start, length);

	int i;
	for (i = start; this->buffer[i] && (this->buffer[i] != c); i++);
	return this->buffer[i] ? i : -1;
}

int mstring::find_dif(char c, int start) const
{
	if (start < 0) start = max(0, this->length + start);
	start = min(start, length);

	int i;
	for (i = start; this->buffer[i] && (this->buffer[i] == c); i++);
	return this->buffer[i] ? i : -1;
}

int mstring::rfind(char c, int start) const
{
	if (start < 0) start = max(0, this->length + start);
	start = min(start, length - 1);

	int i;
	for (i = start; i >= 0 && this->buffer[i] != c; i--);
	return i >= 0 ? i : -1;
}

int mstring::find(const char *str, int start) const
{
	if (start < 0) start = max(0, this->length + start);
	start = min(start, length);

	char *p = strstr(this->buffer + start, str);
	return p ? int(p - this->buffer) : -1;
}

int mstring::find_any(const char *chars, int start) const
{
	if (start < 0) start = max(0, this->length + start);
	start = min(start, length);

	char *p = strpbrk(this->buffer + start, chars);
	return p ? int(p - this->buffer) : -1;
}

mstring mstring::replace(const char *from, const char *to) const
{
	int size_from = (int) strlen(from);
	int size_to = (int) strlen(to);
	int count = 0;

	for (char *pi = this->buffer - size_from; (pi = strstr(pi + size_from, from)) != 0; count++);
	int ret_size = this->length - count * (size_from - size_to);

	int ret_capacity;
	char *ret_buffer = memory::mem_allocate(ret_size + 1, &ret_capacity);
	
	int spos = 0, dpos = 0;
	char *p;

	while ((p = strstr(this->buffer + spos, from)) != 0)
	{
		int n = int(p - this->buffer) - spos;
		if (n > 0)
		{
			strncpy(ret_buffer + dpos, this->buffer + spos, n);
			dpos += n;
			spos += n;
		}
		strcpy(ret_buffer + dpos, to);
		dpos += size_to;
		spos += size_from;
	}

	if (spos < this->length)
	{
		strncpy(ret_buffer + dpos, this->buffer + spos, this->length - spos);
		dpos += this->length - spos;
	}

	ret_buffer[dpos] = 0;

	return mstring(ret_buffer, ret_size, ret_capacity);
}

mstring &mstring::append(const char *str)
{
	//fork();
	if (str)
	{
		int count = (int) strlen(str);
		//if ((this->length + count) >= this->capacity) 
		fork(this->length + count + 1);
		strcpy(this->buffer + this->length, str);
		this->length += count;
	}
	return *this;
}

mstring &mstring::append(const mstring &str)
{
	int count = str.length;
	//if ((this->length + count) >= this->capacity) 
	fork(this->length + count + 1);
	strcpy(this->buffer + this->length, str.buffer);
	this->length += count;
	return *this;
}

mstring& mstring::append(int i)
{
	static char buf[16];
	char *b = &buf[15];
	*b-- = 0;
	
	if (i == 0) return append('0');
	
	if (i < 0) 
	{ 
		append('-'); 
		i = -i; 
	}
	while (i > 0)
	{
		*b-- = char((i % 10) + '0'); 
		i /= 10; 
	}

	return append((const char *)(b+1));
}

mstring &mstring::append(char c)
{
	//if ((this->length + 1) >= this->capacity) 
	fork(this->length + 1 + 1);
	this->buffer[this->length++] = c;
	this->buffer[this->length] = 0;
	return *this;
}

mstring &mstring::set(const char *str)
{
	memory::mem_free(this->buffer);
	if (str) this->length = (int) strlen(str); else this->length = 0;
	this->buffer = memory::mem_allocate(this->length + 1, &this->capacity);
	if (str) strcpy(this->buffer, str); else this->buffer[0] = 0;
	return *this;
}

mstring &mstring::set(const mstring &str)
{
	if (&str == this) return *this;
	memory::mem_free(this->buffer);
	this->length = str.length;
	this->buffer = memory::mem_allocate(this->length + 1, &this->capacity);
	strcpy(this->buffer, str.buffer);
	return *this;
}

/*
void mstring::take(const mstring& str)
{
	if (&str == this) return;
	
	memory::mem_free(this->buffer);
	
	this->buffer = str.buffer;
	this->length = str.length;
	this->capacity = str.capacity;
	
	memory::mem_add_owner(this->buffer);
}

mstring& mstring::operator<<(const fast &b)
{
	mstring ret = b;
	this->take(ret);
	return *this;
}
*/

//mstring& 
void mstring::operator+=(const fast &b)
{
	fork(this->length + b.size() + 1);
	char *p = this->buffer + this->length;
	this->length += b.size();
	p[b.size()] = 0;
	b.write(p);
	//return *this;
}

bool mstring::cstr_equal(const char *a, const char *b)
{
	return strcmp(a, b) == 0;
}

int mstring::cstr_size(const char *s)
{
	return (int) strlen(s);
}

void mstring::cstr_strncpy(char *dest, const char *source, int count)
{
	strncpy(dest, source, count);
}
