#ifndef MSTRING_H
#define MSTRING_H

#include "mstring.h"

class fast;

class mstring
{
	enum { MIN_SIZE = 16 };
	
	char *buffer;
	int length, capacity;
	
	mstring(char *b, int s, int c): buffer(b), length(s), capacity(c) {}
	void fork(int n);
public:
	mstring(const char *str, int capacity = 0);
	mstring(const mstring &str, int capacity = 0);
	explicit mstring(int capacity = MIN_SIZE);
	explicit mstring(char c, int count);
	~mstring();
	
	bool operator==(const mstring &str) const;
	bool operator!=(const mstring &str) const;
	bool operator==(const char * str) const;
	bool operator!=(const char * str) const;
	char operator[](int index) const { return this->buffer[index]; }
	
	operator const char*() const { return buffer; }
	const char *cstr() const { return buffer; }
	bool empty() const { return this->length == 0; }
	int size() const { return this->length; }

	void clear() { fork(1); this->length = 0; this->buffer[0] = 0; }
	
	int find(char c, int start = 0) const;
	int find_dif(char c, int start = 0) const;
	int rfind(char c, int start = -1) const;
	int find(const char *mstring, int start = 0) const;
	int find_any(const char *chars, int start = 0) const;
	
	mstring& append(const mstring &str);
	mstring& append(const char *str);
	mstring& append(char c);
	mstring& append(int i);
	
	mstring replace(const char *from, const char *to) const;
	mstring trim() const;
	mstring substr(int start, int count) const;
	mstring substr(int start) const;
	
	mstring& set(const char *str);
	mstring& set(const mstring &str);
	
	void operator=(const char *str) { set(str); }
	void operator=(const mstring &str) { set(str); }
	
	mstring& operator+=(const char *str) { return append(str); }
	mstring& operator+=(char c) { return append(c); }
	mstring& operator+=(int i) { return append(i); }
	mstring& operator+=(mstring &str) { return append(str); }
	
	mstring& operator<<(const mstring& b) { return append(b); }
	mstring& operator<<(const char *b) { return append(b); }
	mstring& operator<<(char b) { return append(b); }
	mstring& operator<<(int b) { return append(b); }

	void operator+=(const fast &b);

	static bool cstr_equal(const char *a, const char *b);
	static int cstr_size(const char *s);
	static void cstr_strncpy(char *dest, const char *source, int count);
	
	friend class fast;
};

class fast
{
	static const char *ptr_from_mstring(const mstring *a) { return a->buffer; }
	
	struct conc
	{
		enum { t_str, t_charp, t_fast, t_char } type; 
		int size;
		union 
		{
			const mstring *p_str;
			const char *p_charp;
			const fast* p_fast;
			char p_char;
		} u;
		
		conc(const mstring& a): type(t_str) { u.p_str = &a; size = a.size(); }
		conc(const fast& a): type(t_fast) { size = a.d1.size + a.d2.size; u.p_fast = &a; }
		conc(const char *a): type(t_charp) { size = mstring::cstr_size(a); u.p_charp = a; }
		conc(char a): type(t_char) { size = 1; u.p_char = a; }

		void write(char *&p) const
		{
			switch(type)
			{
				case t_str: mstring::cstr_strncpy(p, fast::ptr_from_mstring(u.p_str), size); p += size; break;
				case t_fast : u.p_fast->write(p); break;
				case t_charp: mstring::cstr_strncpy(p, u.p_charp, size); p += size; break;
				case t_char : *p++ = u.p_char; break;  
			}
		}
	} d1, d2;
	
	void write(char * &p) const
	{
		d1.write(p);
		d2.write(p);
	}
	
	int size() const { return d1.size + d2.size; }
	
public:
	
	fast(const mstring& a, const mstring &b): d1(a), d2(b) {}
	fast(const mstring& a, const char *b): d1(a), d2(b) {}
	fast(const mstring& a, char b): d1(a), d2(b) {}
	fast(const mstring& a, const fast& b): d1(a), d2(b) {}
	
	fast(const char *a, const mstring &b): d1(a), d2(b) {}
	fast(const char *a, const char *b): d1(a), d2(b) {}
	fast(const char *a, char b): d1(a), d2(b) {}
	fast(const char *a, const fast& b): d1(a), d2(b) {}
	
	fast(const fast &a, const char *b): d1(a), d2(b) {}
	fast(const fast &a, const mstring& b): d1(a), d2(b) {}
	fast(const fast &a, char b): d1(a), d2(b) {}
	fast(const fast &a, const fast& b): d1(a), d2(b) {}
	
	fast(char a, const char *b): d1(a), d2(b) {}
	fast(char a, const mstring& b): d1(a), d2(b) {}
	fast(char a, char b): d1(a), d2(b) {}
	fast(char a, const fast& b): d1(a), d2(b) {}
	
	operator mstring() const
	{
		mstring ret(size()+1);
		ret.length = size();
		
		char *p = ret.buffer;
		p[size()] = 0;
		write(p);
		
		return ret;
	}
	
	friend class mstring;
};

inline fast const operator+(const mstring& a, const mstring& b) { return fast(a, b); }
inline fast const operator+(const mstring& a, const char* b) { return fast(a, b); }
inline fast const operator+(const mstring& a, char b) { return fast(a, b); }
inline fast const operator+(const mstring& a, const fast& b) { return fast(a, b); }

inline fast const operator+(const char *a, const mstring& b) { return fast(a, b); }
inline fast const operator+(const char *a, const fast& b) { return fast(a, b); }

inline fast const operator+(const fast& a, const char * b) { return fast(a, b); }
inline fast const operator+(const fast& a, const mstring& b) { return fast(a, b); }
inline fast const operator+(const fast& a, char b) { return fast(a, b); }
inline fast const operator+(const fast& a, const fast& b) { return fast(a, b); }

inline fast const operator+(char a, const mstring& b) { return fast(a, b); }
inline fast const operator+(char a, const fast& b) { return fast(a, b); }

#endif
