#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "mstring.h"
#include "types.h"
#include "cpu.h"
#include "sound.h"
#include "io.h"
#include "forms.h"
#include "pce.h"
#include "video.h"


int sound_volume = 100;
gp2x::video::surface8 *surf = 0, *surf_pce = 0;
static byte *ROM = 0;


int load_game(const char *name)
{
	const int PAGE_SIZE = 0x2000;
	const int PAGE_SPACE = 0x2008;

	FILE *fp = fopen(name, "rb");

	if (!fp) return 0;

	fseek(fp, 0, SEEK_END);

	int fsize = ftell(fp);
	int rsize = fsize >> 13;

	fseek(fp, fsize & 0x1fff, SEEK_SET);

	ROM = (byte *) malloc(rsize * PAGE_SPACE);

	for (int i = 0; i < rsize; i++) fread(ROM + i * PAGE_SPACE, 1, PAGE_SIZE, fp);

	fclose(fp);

	new_rom(ROM, rsize); // automatically resets the cpu
	//reset_sound();

	return 1;
}

int init_pce()
{
	init_cpu();
	init_video();
	init_io();
	init_sound();

	return 1;
}

int reset_pce()
{
	reset_video();
	reset_sound();
	reset_io();
	if (ROM)
	{
		free(ROM);
		ROM = 0;
	}

	return 1;
}

void trash_pce()
{
	//machine::sound_play(false);
	trash_cpu();
	trash_video();
	if (ROM) free(ROM);
}

static inline int min(const int a, const int b)
{
	return a <= b ? a : b;
}

int run_game(const char *name)
{
	const int FRAMES_PER_SECOND = 60;
	const int SCANLINE_CYCLES = (BASE_CLOCK / SCANLINES_PER_FRAME) / FRAMES_PER_SECOND;
	const int TIMER_TICKS = BASE_CLOCK / 1000 * 3 * 1024 / 21480;

	if (!load_game(name)) return 0;

	int scanline_cycles = SCANLINE_CYCLES; // 453
	int timer_cycles = TIMER_TICKS;
	int ok = 1;

	bool sound_started = false;

	surf->clear(true);
	surf->flip();

	//for (int ok = 1; ok; )
	while (1)
	{
		int todo = min(scanline_cycles, timer_cycles);
		//if (todo <= 0) todo = 1;

		if (todo > 0)
		{
			int done = cpu_cycles_counter;
			//run_cpu(todo);
			//do
			//{
				run_cpu(todo);
				//if (!transfer_pending) break;
				//todo = 5;
			//}
			//while (1);
			done = cpu_cycles_counter - done;

			scanline_cycles -= done;
			timer_cycles -= done;
		}

		while (scanline_cycles <= 0 && ok)
		//if (scanline_cycles <= 0 /* && ok */)
		{
			ok = draw_scanline();
			//draw_scanline();
			scanline_cycles += SCANLINE_CYCLES;
		}

      	if (!ok) break;

		while (timer_cycles <= 0)
      	//if (timer_cycles <= 0)
      	{
      		timer_tick();
      		timer_cycles += TIMER_TICKS;
      	}

      	if (!sound_started)
      	{
      		sound_started = true;
      		gp2x::sound::play();
      	}
	}

	if (sound_started) gp2x::sound::pause();

	return 1;
}


int main()
{
	//if (machine::init(44100, SOUND_BUF_SIZE, sound_producer))
	{
		gp2x::sound::setup(44100, true, sound_producer);

		surf = new gp2x::video::surface8(WIDTH, HEIGHT, true, true);
		surf_pce = new gp2x::video::surface8(WIDTH, HEIGHT, false, false);

		if (init_pce())
		{
			do
			{
				mstring filename = forms::DialogFileOpen("/mnt/sd/roms/pcengine/");
				if (filename.empty()) break;

				if (reset_pce()) run_game(filename); else break;
				//machine::sound_play(false);
			}
			while (1);
		}
		trash_pce();

		if (surf_pce) delete surf_pce;
		if (surf) delete surf;
	}

	//machine::trash();

	return 0;
}
