#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "types.h"
#include "io.h"
#include "video.h"
#include "sound.h"
#include "cpu.h"
#include "gp2x.h"

#define	VDC_CR	0x01
#define	VDC_OR	0x02
#define	VDC_RR	0x04

#define	VDC_BSY	0x40
#define	VDC_SpHit	VDC_CR
#define	VDC_Over	VDC_OR
#define	VDC_RasHit	VDC_RR
#define	VDC_InVBlank VDC_VD

#define	SpHitON		(io.VDC[CR].W&0x01)
#define	OverON		(io.VDC[CR].W&0x02)
#define	RasHitON	(io.VDC[CR].W&0x04)
#define	VBlankON	(io.VDC[CR].W&0x08)
#define	SpriteON	(io.VDC[CR].W&0x40)
#define	ScreenON	(io.VDC[CR].W&0x80)

#define SATBIntON	(io.VDC[DCR].W&0x01)
#define DMAIntON	(io.VDC[DCR].W&0x02)

#define	IRQ1	2

int ScrollYDiff;
int	oldScrollX;
int	oldScrollY;
int	oldScrollYDiff;
int scroll = 0;

int scanline, prevline;

typedef struct {
	short y,x,no,atr;
} SPR;

#define	ZW	64
byte *VRAM,*ROM,*vchange,*WRAM;
short SPRAM[64*4];
unsigned long *VRAM2;
byte *vchanges;
unsigned long *VRAMS;
#define MAXDISP	227
byte BGONSwitch = 1, SPONSwitch = 1;

//byte XBuf[HEIGHT * WIDTH];

static byte *XBuf = 0;

byte Pal[512];

bool scale_to_fit = true;

static inline int min(int a, int b)
{
	return a <= b ? a : b;
}

int CheckSprites(void)
{
	int i,x0,y0,w0,h0,x,y,w,h;
	SPR *spr;
	spr = (SPR*)SPRAM;
	x0 = spr->x;
	y0 = spr->y;
	w0 = (((spr->atr>>8 )&1)+1)*16;
	h0 = (((spr->atr>>12)&3)+1)*16;
	spr++;
	for(i=1;i<64;i++,spr++) {
		x = spr->x;
		y = spr->y;
		w = (((spr->atr>>8 )&1)+1)*16;
		h = (((spr->atr>>12)&3)+1)*16;
		if ((x<x0+w0)&&(x+w>x0)&&(y<y0+h0)&&(y+h>y0)) return 1;
	}
	return 0;
}

static void plane2pixel(int no)
{
	unsigned long M;
	byte *C=VRAM+no*32;
	unsigned long L,*C2 = VRAM2+no*8;
	int j;
	for(j=0;j<8;j++,C+=2,C2++) {
		M=C[0];
		L =((M&0x88)>>3)|((M&0x44)<<6)|((M&0x22)<<15)|((M&0x11)<<24);
		M=C[1];
		L|=((M&0x88)>>2)|((M&0x44)<<7)|((M&0x22)<<16)|((M&0x11)<<25);
		M=C[16];
		L|=((M&0x88)>>1)|((M&0x44)<<8)|((M&0x22)<<17)|((M&0x11)<<26);
		M=C[17];
		L|=((M&0x88))|((M&0x44)<<9)|((M&0x22)<<18)|((M&0x11)<<27);
		C2[0] = L; //37261504
	}
}

static void sprite2pixel(int no)
{
	byte M;
	byte *C;
	unsigned long *C2;
	int i;

	C=&VRAM[no*128];
	C2=&VRAMS[no*32];
	for(i=0;i<32;i++,C++,C2++){
		long L;
		M=C[0];
		L =((M&0x88)>>3)|((M&0x44)<<6)|((M&0x22)<<15)|((M&0x11)<<24);
		M=C[32];
		L|=((M&0x88)>>2)|((M&0x44)<<7)|((M&0x22)<<16)|((M&0x11)<<25);
		M=C[64];
		L|=((M&0x88)>>1)|((M&0x44)<<8)|((M&0x22)<<17)|((M&0x11)<<26);
		M=C[96];
		L|=((M&0x88))|((M&0x44)<<9)|((M&0x22)<<18)|((M&0x11)<<27);
		C2[0]=L;
	}
}

#define	PAL(c)	R[c]
#define	SPal	(Pal+256)
#define	FC_W	io.screen_w
#define	FC_H	256

void RefreshLine(int Y1,int Y2)
{
	int X1,XW,Line;
	int x,y,h,offset;

	byte *PP;
	Y2++;
	PP=XBuf+WIDTH*(HEIGHT-FC_H)/2+(WIDTH-FC_W)/2+WIDTH*Y1;
	if(!ScreenON || !BGONSwitch)
		memset(XBuf+((HEIGHT-FC_H)/2+Y1)*WIDTH,Pal[0],(Y2-Y1)*WIDTH);
	else {

		y = Y1+ScrollY-ScrollYDiff;
		offset = y&7;
		h = 8-offset;
		if (h>Y2-Y1) h=Y2-Y1;
		y>>=3;
		PP-=ScrollX&7;
		XW=io.screen_w/8+1;

		for(Line=Y1;Line<Y2;y++) {
			x = ScrollX/8;
			y &= io.bg_h-1;
			for(X1=0;X1<XW;X1++,x++,PP+=8/*,ZP++*/){
				byte *R,*P,*C;//,*Z;
				unsigned long *C2;
				int no,i;
				x&=io.bg_w-1;
				no = ((word*)VRAM)[x+y*io.bg_w];
				R = &Pal[(no>>12)*16];
				no&=0xFFF;
				if (vchange[no]) { vchange[no]=0; plane2pixel(no);}
				C2 = &VRAM2[no*8+offset];
				C = &VRAM[no*32+offset*2];
				P = PP;
				for(i=0;i<h;i++,P+=WIDTH,C2++,C+=2/*,Z+=ZW*/) {
					unsigned long L;

					L=C2[0];
					P[0]=PAL((L>>4)&15);
					P[1]=PAL((L>>12)&15);
					P[2]=PAL((L>>20)&15);
					P[3]=PAL((L>>28));
					P[4]=PAL((L)&15);
					P[5]=PAL((L>>8)&15);
					P[6]=PAL((L>>16)&15);
					P[7]=PAL((L>>24)&15);
				}
			}
			Line+=h;
			PP+=WIDTH*h-XW*8;
			offset = 0;
			h = Y2-Line;
			if (h>8) h=8;
		}
	}
}

#define	V_FLIP	0x8000
#define	H_FLIP	0x0800
#define	SPBG	0x80
#define	CGX		0x100

static void PutSprite(byte *P,byte *C,unsigned long *C2,byte *R,int h,int inc)
{
	int i;
	word J;
	unsigned long L;
	for(i=0;i<h;i++,C+=inc,C2+=inc,P+=WIDTH) {
		J = ((word*)C)[0]|((word*)C)[16]|((word*)C)[32]|((word*)C)[48];
		if (!J) continue;
		L = C2[1];
		if (J&0x8000) P[0]=PAL((L>>4)&15);
		if (J&0x4000) P[1]=PAL((L>>12)&15);
		if (J&0x2000) P[2]=PAL((L>>20)&15);
		if (J&0x1000) P[3]=PAL((L>>28));
		if (J&0x0800) P[4]=PAL((L)&15);
		if (J&0x0400) P[5]=PAL((L>>8)&15);
		if (J&0x0200) P[6]=PAL((L>>16)&15);
		if (J&0x0100) P[7]=PAL((L>>24)&15);
		L = C2[0];
		if (J&0x80) P[8 ]=PAL((L>>4)&15);
		if (J&0x40) P[9 ]=PAL((L>>12)&15);
		if (J&0x20) P[10]=PAL((L>>20)&15);
		if (J&0x10) P[11]=PAL((L>>28));
		if (J&0x08) P[12]=PAL((L)&15);
		if (J&0x04) P[13]=PAL((L>>8)&15);
		if (J&0x02) P[14]=PAL((L>>16)&15);
		if (J&0x01) P[15]=PAL((L>>24)&15);
	}
}

static void PutSpriteHflip(byte *P,byte *C,unsigned long *C2,byte *R,int h,int inc)
{
	int i;
	word J;
	unsigned long L;
	for(i=0;i<h;i++,C+=inc,C2+=inc,P+=WIDTH) {
		J = ((word*)C)[0]|((word*)C)[16]|((word*)C)[32]|((word*)C)[48];
		if (!J) continue;
		L = C2[0];
		if (J&0x01) P[0]=PAL((L>>24)&15);
		if (J&0x02) P[1]=PAL((L>>16)&15);
		if (J&0x04) P[2]=PAL((L>>8)&15);
		if (J&0x08) P[3]=PAL((L)&15);
		if (J&0x10) P[4]=PAL((L>>28));
		if (J&0x20) P[5]=PAL((L>>20)&15);
		if (J&0x40) P[6]=PAL((L>>12)&15);
		if (J&0x80) P[7]=PAL((L>>4)&15);
		L = C2[1];
		if (J&0x0100) P[8]=PAL((L>>24)&15);
		if (J&0x0200) P[9]=PAL((L>>16)&15);
		if (J&0x0400) P[10]=PAL((L>>8)&15);
		if (J&0x0800) P[11]=PAL((L)&15);
		if (J&0x1000) P[12]=PAL((L>>28));
		if (J&0x2000) P[13]=PAL((L>>20)&15);
		if (J&0x4000) P[14]=PAL((L>>12)&15);
		if (J&0x8000) P[15]=PAL((L>>4)&15);
	}
}

static void PutSpriteBack(byte *P,byte *C,unsigned long *C2,byte *R,int h,int inc,
	int cx,int cy,int xoffset,int yoffset)
{
	int i;
	word J;
	unsigned long L;
	unsigned long M;
	int tx;
	byte *CC1,*CC2,*CC3;
	int no;

	yoffset+=8;
	cy--;
	for(i=0;i<h;i++,C+=inc,C2+=inc,P+=WIDTH,yoffset++,CC1+=2,CC2+=2,CC3+=2) {
		if (yoffset>=8) {
			cy++;
			cy = cy&(io.bg_h-1);
			yoffset -= 8;

			no = ((word*)VRAM)[cx+cy*io.bg_w];
			no&=0xFFF;
			CC1 = &VRAM[no*32+yoffset*2];
			tx=(cx+1)&(io.bg_w-1);
			no = ((word*)VRAM)[tx+cy*io.bg_w];
			no&=0xFFF;
			CC2 = &VRAM[no*32+yoffset*2];
			tx=(tx+1)&(io.bg_w-1);
			no = ((word*)VRAM)[tx+cy*io.bg_w];
			no&=0xFFF;
			CC3 = &VRAM[no*32+yoffset*2];
		}

		M=CC1[0]|CC1[1]|CC1[16]|CC1[17];
		M=(M<<8)|CC2[0]|CC2[1]|CC2[16]|CC2[17];
		M=(M<<8)|CC3[0]|CC3[1]|CC3[16]|CC3[17];
		M>>=(8-xoffset);
		M=~M;

		J = ((word*)C)[0]|((word*)C)[16]|((word*)C)[32]|((word*)C)[48];
		if (!J) continue;
		L = C2[1];
		if (J&M&0x8000) P[0]=PAL((L>>4)&15);
		if (J&M&0x4000) P[1]=PAL((L>>12)&15);
		if (J&M&0x2000) P[2]=PAL((L>>20)&15);
		if (J&M&0x1000) P[3]=PAL((L>>28));
		if (J&M&0x0800) P[4]=PAL((L)&15);
		if (J&M&0x0400) P[5]=PAL((L>>8)&15);
		if (J&M&0x0200) P[6]=PAL((L>>16)&15);
		if (J&M&0x0100) P[7]=PAL((L>>24)&15);
		L = C2[0];
		if (J&M&0x80) P[8 ]=PAL((L>>4)&15);
		if (J&M&0x40) P[9 ]=PAL((L>>12)&15);
		if (J&M&0x20) P[10]=PAL((L>>20)&15);
		if (J&M&0x10) P[11]=PAL((L>>28));
		if (J&M&0x08) P[12]=PAL((L)&15);
		if (J&M&0x04) P[13]=PAL((L>>8)&15);
		if (J&M&0x02) P[14]=PAL((L>>16)&15);
		if (J&M&0x01) P[15]=PAL((L>>24)&15);
	}
}

static void PutSpriteHflipBack(byte *P,byte *C,unsigned long *C2,byte *R,int h,int inc,
	int cx,int cy,int xoffset,int yoffset)
{
	int i;
	word J;
	unsigned long L;
	unsigned long M;
	int tx;
	byte *CC1,*CC2,*CC3;
	int no;

	yoffset+=8;
	cy--;
	for(i=0;i<h;i++,C+=inc,C2+=inc,P+=WIDTH,yoffset++,CC1+=2,CC2+=2,CC3+=2) {
		if (yoffset>=8) {
			cy++;
			cy = cy&(io.bg_h-1);
			yoffset -= 8;

			no = ((word*)VRAM)[cx+cy*io.bg_w];
			no&=0xFFF;
			CC1 = &VRAM[no*32+yoffset*2];
			tx=(cx+1)&(io.bg_w-1);
			no = ((word*)VRAM)[tx+cy*io.bg_w];
			no&=0xFFF;
			CC2 = &VRAM[no*32+yoffset*2];
			tx=(tx+1)&(io.bg_w-1);
			no = ((word*)VRAM)[tx+cy*io.bg_w];
			no&=0xFFF;
			CC3 = &VRAM[no*32+yoffset*2];
		}

		M=CC1[0]|CC1[1]|CC1[16]|CC1[17];
		M=(M<<8)|CC2[0]|CC2[1]|CC2[16]|CC2[17];
		M=(M<<8)|CC3[0]|CC3[1]|CC3[16]|CC3[17];
		M>>=(8-xoffset);

		J = ((word*)C)[0]|((word*)C)[16]|((word*)C)[32]|((word*)C)[48];
		if (!J) continue;
		L = C2[0];
		if ((J&0x01|M&0x8000)==0x01) P[0]=PAL((L>>24)&15);
		if ((J&0x02|M&0x4000)==0x02) P[1]=PAL((L>>16)&15);
		if ((J&0x04|M&0x2000)==0x04) P[2]=PAL((L>>8)&15);
		if ((J&0x08|M&0x1000)==0x08) P[3]=PAL((L)&15);
		if ((J&0x10|M&0x0800)==0x10) P[4]=PAL((L>>28));
		if ((J&0x20|M&0x0400)==0x20) P[5]=PAL((L>>20)&15);
		if ((J&0x40|M&0x0200)==0x40) P[6]=PAL((L>>12)&15);
		if ((J&0x80|M&0x0100)==0x80) P[7]=PAL((L>>4)&15);
		L = C2[1];
		if ((J&0x0100|M&0x80)==0x0100) P[8]=PAL((L>>24)&15);
		if ((J&0x0200|M&0x40)==0x0200) P[9]=PAL((L>>16)&15);
		if ((J&0x0400|M&0x20)==0x0400) P[10]=PAL((L>>8)&15);
		if ((J&0x0800|M&0x10)==0x0800) P[11]=PAL((L)&15);
		if ((J&0x1000|M&0x08)==0x1000) P[12]=PAL((L>>28));
		if ((J&0x2000|M&0x04)==0x2000) P[13]=PAL((L>>20)&15);
		if ((J&0x4000|M&0x02)==0x4000) P[14]=PAL((L>>12)&15);
		if ((J&0x8000|M&0x01)==0x8000) P[15]=PAL((L>>4)&15);
	}
}

static void PutSpriteBack2(byte *P,byte *C,unsigned long *C2,byte *R,int h,int inc,
	int cx,int cy,int xoffset,int yoffset)
{
	int i;
	unsigned long J;
	unsigned long L;
	unsigned long M;
	unsigned long CL;
	int tx;
	byte *CC1,*CC2,*CC3;
	unsigned long *CC21,*CC22,*CC23;
	byte *CR1, *CR2, *CR3;
	byte *CP;
	int no;

	yoffset+=8;
	cy--;
	for(i=0;i<h;i++,C+=inc,C2+=inc,P+=WIDTH,yoffset++,CC1+=2,CC2+=2,CC3+=2,CC21++,CC22++,CC23++) {
		if (yoffset>=8) {
			cy++;
			cy = cy&(io.bg_h-1);
			yoffset -= 8;

			no = ((word*)VRAM)[cx+cy*io.bg_w];
			CR1 = &Pal[(no>>12)*16];
			no&=0xFFF;
			CC1 = &VRAM[no*32+yoffset*2];
			CC21 = &VRAM2[no*8+yoffset];
			tx=(cx+1)&(io.bg_w-1);
			no = ((word*)VRAM)[tx+cy*io.bg_w];
			CR2 = &Pal[(no>>12)*16];
			no&=0xFFF;
			CC2 = &VRAM[no*32+yoffset*2];
			CC22 = &VRAM2[no*8+yoffset];
			tx=(tx+1)&(io.bg_w-1);
			no = ((word*)VRAM)[tx+cy*io.bg_w];
			CR3 = &Pal[(no>>12)*16];
			no&=0xFFF;
			CC3 = &VRAM[no*32+yoffset*2];
			CC23 = &VRAM2[no*8+yoffset];
		}

		M=CC1[0]|CC1[1]|CC1[16]|CC1[17];
		M=(M<<8)|CC2[0]|CC2[1]|CC2[16]|CC2[17];
		M=(M<<8)|CC3[0]|CC3[1]|CC3[16]|CC3[17];
		M>>=(8-xoffset);
		M=~M;

		J = ((word*)C)[0]|((word*)C)[16]|((word*)C)[32]|((word*)C)[48];
		if (!J) continue;
		L = C2[1];
		if (J&M&0x8000) P[0]=PAL((L>>4)&15);
		if (J&M&0x4000) P[1]=PAL((L>>12)&15);
		if (J&M&0x2000) P[2]=PAL((L>>20)&15);
		if (J&M&0x1000) P[3]=PAL((L>>28));
		if (J&M&0x0800) P[4]=PAL((L)&15);
		if (J&M&0x0400) P[5]=PAL((L>>8)&15);
		if (J&M&0x0200) P[6]=PAL((L>>16)&15);
		if (J&M&0x0100) P[7]=PAL((L>>24)&15);
		L = C2[0];
		if (J&M&0x80) P[8 ]=PAL((L>>4)&15);
		if (J&M&0x40) P[9 ]=PAL((L>>12)&15);
		if (J&M&0x20) P[10]=PAL((L>>20)&15);
		if (J&M&0x10) P[11]=PAL((L>>28));
		if (J&M&0x08) P[12]=PAL((L)&15);
		if (J&M&0x04) P[13]=PAL((L>>8)&15);
		if (J&M&0x02) P[14]=PAL((L>>16)&15);
		if (J&M&0x01) P[15]=PAL((L>>24)&15);

		M<<=8-xoffset;
		M=~M;
		J<<=8-xoffset;
		CL=CC21[0];
		CP=P-xoffset;
		if (xoffset<=0&&(J&M&0x800000)) CP[0]=CR1[(CL>>4)&15];
		if (xoffset<=1&&(J&M&0x400000)) CP[1]=CR1[(CL>>12)&15];
		if (xoffset<=2&&(J&M&0x200000)) CP[2]=CR1[(CL>>20)&15];
		if (xoffset<=3&&(J&M&0x100000)) CP[3]=CR1[(CL>>28)];
		if (xoffset<=4&&(J&M&0x080000)) CP[4]=CR1[(CL)&15];
		if (xoffset<=5&&(J&M&0x040000)) CP[5]=CR1[(CL>>8)&15];
		if (xoffset<=6&&(J&M&0x020000)) CP[6]=CR1[(CL>>16)&15];
		if (            (J&M&0x010000)) CP[7]=CR1[(CL>>24)&15];
		CL=CC22[0];
		if (J&M&0x8000) CP[8 ]=CR2[(CL>>4)&15];
		if (J&M&0x4000) CP[9 ]=CR2[(CL>>12)&15];
		if (J&M&0x2000) CP[10]=CR2[(CL>>20)&15];
		if (J&M&0x1000) CP[11]=CR2[(CL>>28)];
		if (J&M&0x0800) CP[12]=CR2[(CL)&15];
		if (J&M&0x0400) CP[13]=CR2[(CL>>8)&15];
		if (J&M&0x0200) CP[14]=CR2[(CL>>16)&15];
		if (J&M&0x0100) CP[15]=CR2[(CL>>24)&15];
		CL=CC23[0];
		if (xoffset>=1&&(J&M&0x80)) CP[16]=CR3[(CL>>4)&15];
		if (xoffset>=2&&(J&M&0x40)) CP[17]=CR3[(CL>>12)&15];
		if (xoffset>=3&&(J&M&0x20)) CP[18]=CR3[(CL>>20)&15];
		if (xoffset>=4&&(J&M&0x10)) CP[19]=CR3[(CL>>28)];
		if (xoffset>=5&&(J&M&0x08)) CP[20]=CR3[(CL)&15];
		if (xoffset>=6&&(J&M&0x04)) CP[21]=CR3[(CL>>8)&15];
		if (xoffset>=7&&(J&M&0x02)) CP[22]=CR3[(CL>>16)&15];
	}
}

static void PutSpriteHflipBack2(byte *P,byte *C,unsigned long *C2,byte *R,int h,int inc,
	int cx,int cy,int xoffset,int yoffset)
{
	int i;
	unsigned long J;
	unsigned long L;
	unsigned long M;
	unsigned long CL;
	int tx;
	byte *CC1,*CC2,*CC3;
	unsigned long *CC21,*CC22,*CC23;
	byte *CR1, *CR2, *CR3;
	byte *CP;
	int no;

	yoffset+=8;
	cy--;
	for(i=0;i<h;i++,C+=inc,C2+=inc,P+=WIDTH,yoffset++,CC1+=2,CC2+=2,CC3+=2,CC21++,CC22++,CC23++) {
		if (yoffset>=8) {
			cy++;
			cy = cy&(io.bg_h-1);
			yoffset -= 8;

			no = ((word*)VRAM)[cx+cy*io.bg_w];
			CR1 = &Pal[(no>>12)*16];
			no&=0xFFF;
			CC1 = &VRAM[no*32+yoffset*2];
			CC21 = &VRAM2[no*8+yoffset];
			tx=(cx+1)&(io.bg_w-1);
			no = ((word*)VRAM)[tx+cy*io.bg_w];
			CR2 = &Pal[(no>>12)*16];
			no&=0xFFF;
			CC2 = &VRAM[no*32+yoffset*2];
			CC22 = &VRAM2[no*8+yoffset];
			tx=(tx+1)&(io.bg_w-1);
			no = ((word*)VRAM)[tx+cy*io.bg_w];
			CR3 = &Pal[(no>>12)*16];
			no&=0xFFF;
			CC3 = &VRAM[no*32+yoffset*2];
			CC23 = &VRAM2[no*8+yoffset];
		}

		M=CC1[0]|CC1[1]|CC1[16]|CC1[17];
		M=(M<<8)|CC2[0]|CC2[1]|CC2[16]|CC2[17];
		M=(M<<8)|CC3[0]|CC3[1]|CC3[16]|CC3[17];
		M>>=(8-xoffset);

		J = ((word*)C)[0]|((word*)C)[16]|((word*)C)[32]|((word*)C)[48];
		if (!J) continue;
		L = C2[0];
		if ((J&0x01|M&0x8000)==0x01) P[0]=PAL((L>>24)&15);
		if ((J&0x02|M&0x4000)==0x02) P[1]=PAL((L>>16)&15);
		if ((J&0x04|M&0x2000)==0x04) P[2]=PAL((L>>8)&15);
		if ((J&0x08|M&0x1000)==0x08) P[3]=PAL((L)&15);
		if ((J&0x10|M&0x0800)==0x10) P[4]=PAL((L>>28));
		if ((J&0x20|M&0x0400)==0x20) P[5]=PAL((L>>20)&15);
		if ((J&0x40|M&0x0200)==0x40) P[6]=PAL((L>>12)&15);
		if ((J&0x80|M&0x0100)==0x80) P[7]=PAL((L>>4)&15);
		L = C2[1];
		if ((J&0x0100|M&0x80)==0x0100) P[8]=PAL((L>>24)&15);
		if ((J&0x0200|M&0x40)==0x0200) P[9]=PAL((L>>16)&15);
		if ((J&0x0400|M&0x20)==0x0400) P[10]=PAL((L>>8)&15);
		if ((J&0x0800|M&0x10)==0x0800) P[11]=PAL((L)&15);
		if ((J&0x1000|M&0x08)==0x1000) P[12]=PAL((L>>28));
		if ((J&0x2000|M&0x04)==0x2000) P[13]=PAL((L>>20)&15);
		if ((J&0x4000|M&0x02)==0x4000) P[14]=PAL((L>>12)&15);
		if ((J&0x8000|M&0x01)==0x8000) P[15]=PAL((L>>4)&15);

		M<<=8-xoffset;
		J<<=xoffset;
		CL=CC21[0];
		CP=P-xoffset;
		if (xoffset<=0&&(J&0x01|M&0x800000)==0x800001) CP[0]=CR1[(CL>>4)&15];
		if (xoffset<=1&&(J&0x02|M&0x400000)==0x400002) CP[1]=CR1[(CL>>12)&15];
		if (xoffset<=2&&(J&0x04|M&0x200000)==0x200004) CP[2]=CR1[(CL>>20)&15];
		if (xoffset<=3&&(J&0x08|M&0x100000)==0x100008) CP[3]=CR1[(CL>>28)];
		if (xoffset<=4&&(J&0x10|M&0x080000)==0x080010) CP[4]=CR1[(CL)&15];
		if (xoffset<=5&&(J&0x20|M&0x040000)==0x040020) CP[5]=CR1[(CL>>8)&15];
		if (xoffset<=6&&(J&0x40|M&0x020000)==0x020040) CP[6]=CR1[(CL>>16)&15];
		if (            (J&0x80|M&0x010000)==0x010080) CP[7]=CR1[(CL>>24)&15];
		CL=CC22[0];
		if ((J&0x0100|M&0x8000)==0x8100) CP[8 ]=CR2[(CL>>4)&15];
		if ((J&0x0200|M&0x4000)==0x4200) CP[9 ]=CR2[(CL>>12)&15];
		if ((J&0x0400|M&0x2000)==0x2400) CP[10]=CR2[(CL>>20)&15];
		if ((J&0x0800|M&0x1000)==0x1800) CP[11]=CR2[(CL>>28)];
		if ((J&0x1000|M&0x0800)==0x1800) CP[12]=CR2[(CL)&15];
		if ((J&0x2000|M&0x0400)==0x2400) CP[13]=CR2[(CL>>8)&15];
		if ((J&0x4000|M&0x0200)==0x4200) CP[14]=CR2[(CL>>16)&15];
		if ((J&0x8000|M&0x0100)==0x8100) CP[15]=CR2[(CL>>24)&15];
		CL=CC23[0];
		if (xoffset>=1&&(J&0x010000|M&0x80)==0x010080) CP[16]=CR3[(CL>>4)&15];
		if (xoffset>=2&&(J&0x020000|M&0x40)==0x020040) CP[17]=CR3[(CL>>12)&15];
		if (xoffset>=3&&(J&0x040000|M&0x20)==0x040020) CP[18]=CR3[(CL>>20)&15];
		if (xoffset>=4&&(J&0x080000|M&0x10)==0x080010) CP[19]=CR3[(CL>>28)];
		if (xoffset>=5&&(J&0x100000|M&0x08)==0x100008) CP[20]=CR3[(CL)&15];
		if (xoffset>=6&&(J&0x200000|M&0x04)==0x200004) CP[21]=CR3[(CL>>8)&15];
		if (xoffset>=7&&(J&0x400000|M&0x02)==0x400002) CP[22]=CR3[(CL>>16)&15];
	}
}

void RefreshSprite(int Y1,int Y2)
{
	int n;
	SPR *spr;
	int usespbg = 0;

	spr = (SPR*)SPRAM + 63;

	for(n=0;n<64;n++,spr--){
		int x,y,no,atr,inc,cgx,cgy;
		byte *R,*C;
		unsigned long *C2;
		int	pos;
		int h,t,i,j;
		int y_sum;
		int spbg;
		int cx;
		int cy;
		int yoffset;
		int xoffset;
		int tx;

		atr=spr->atr;
		spbg = (atr>>7)&1;

		y = (spr->y&1023)-64;
		x = (spr->x&1023)-32;
		no= spr->no&2047;
		cgx = (atr>>8)&1;
		cgy = (atr>>12)&3;
		cgy |= cgy>>1;
		no = (no>>1)&~(cgy*2+cgx);
		if (y>=Y2 || y+(cgy+1)*16<Y1 || x>=FC_W || x+(cgx+1)*16<0) continue;

		R = &SPal[(atr&15)*16];

		for (i=0;i<cgy*2+cgx+1;i++) {
			if (vchanges[no+i]) {
				vchanges[no+i] = 0;
				sprite2pixel(no+i);
			}
			if (!cgx) i++;
		}

		C = &VRAM[no*128];
		C2 = &VRAMS[no*32];

		pos = WIDTH*(HEIGHT-FC_H)/2+(WIDTH-FC_W)/2+WIDTH*y+x;
		inc = 2;
		if (atr&V_FLIP) { inc=-2; C+=15*2+cgy*256; C2+=15*2+cgy*64;}
		y_sum = 0;

		cy = y+ScrollY-ScrollYDiff;
		yoffset = cy&7;
		cy>>=3;
		xoffset=(x+ScrollX)&7;
		cx = (x+ScrollX)/8;

		for(i=0;i<=cgy;i++) {
			cy = cy&(io.bg_h-1);
			t = Y1-y-y_sum;
			h = 16;
			if (t>0) {
				C+=t*inc;
				C2+=t*inc;
				h-=t;
				pos+=t*WIDTH;
				cy+=(yoffset+t)>>3;
				yoffset=(yoffset+t)&7;
			}
			if (h>Y2-y-y_sum) h = Y2-y-y_sum;
			if (spbg==1 || !ScreenON){
				usespbg=1;
				if (atr&H_FLIP){
					for(j=0;j<=cgx;j++)
						PutSpriteHflip(XBuf+pos+(cgx-j)*16,C+j*128,C2+j*32,R,h,inc);
				}else{
					for(j=0;j<=cgx;j++)
						PutSprite(XBuf+pos+j*16,C+j*128,C2+j*32,R,h,inc);
				}
			} else if (usespbg) {
				if (atr&H_FLIP){
					for(j=0,tx=cx;j<=cgx;j++,tx+=2) {
						tx&=io.bg_w-1;
						PutSpriteHflipBack2(XBuf+pos+j*16,C+(cgx-j)*128,C2+(cgx-j)*32,R,h,inc,tx,cy,xoffset,yoffset);
					}
				}else{
					for(j=0,tx=cx;j<=cgx;j++,tx+=2) {
						tx&=io.bg_w-1;
						PutSpriteBack2(XBuf+pos+j*16,C+j*128,C2+j*32,R,h,inc,tx,cy,xoffset,yoffset);
					}
				}
			} else {
				if (atr&H_FLIP){
					for(j=0,tx=cx;j<=cgx;j++,tx+=2) {
						tx&=io.bg_w-1;
						PutSpriteHflipBack(XBuf+pos+j*16,C+(cgx-j)*128,C2+(cgx-j)*32,R,h,inc,tx,cy,xoffset,yoffset);
					}
				}else{
					for(j=0,tx=cx;j<=cgx;j++,tx+=2) {
						tx&=io.bg_w-1;
						PutSpriteBack(XBuf+pos+j*16,C+j*128,C2+j*32,R,h,inc,tx,cy,xoffset,yoffset);
					}
				}
			}
			pos+=h*WIDTH;
			C+=h*inc+16*7*inc;
			C2+=h*inc+16*inc;
			y_sum+=16;
			cy+=(yoffset+h)>>3;
			yoffset=(yoffset+h)&7;
		}
	}
}

static inline int max(int a, int b)
{
	return a >= b ? a : b;
}

void RefreshScreen()
{
	int dispmin = (MaxLine-MinLine>MAXDISP ? MinLine+((MaxLine-MinLine-MAXDISP+1)>>1) : MinLine);
	int dispmax = (MaxLine-MinLine>MAXDISP ? MaxLine-((MaxLine-MinLine-MAXDISP+1)>>1) : MaxLine);

	int x = (WIDTH - FC_W) / 2;
	int y = (HEIGHT - FC_H) / 2 + MinLine + dispmin;

	int w = FC_W;
	int h = dispmax - dispmin + 1;

	int extra_x = max(0, (320- w));
	int extra_y = max(0, (240 - h));

	int margin_x = extra_x >> 1;
	int margin_y = extra_y >> 1;


	/*
	surf->clear(false);
	*/

	surf->blit(surf_pce, x, y, w, h, margin_x, margin_y);


	int fps = gp2x::video::sync_fps(60);
	char fps_str[4];

	fps_str[0] = fps / 100 + '0';
	fps_str[1] = (fps / 10) % 10 + '0';
	fps_str[2] = fps % 10 + '0';
	fps_str[3] = 0;


	if (w > 320 || h > 240)
	{
		if (scale_to_fit)
		{
			surf->print(font_7x13, 0, 0, fps_str, 255, 0, true);
			surf->flip(0, 0, w < 320 ? 320 : w, h < 240 ? 240 : h);
			//surf->flip(0, 0, w < 320 ? 320 : w, 240);
		}
		else
		{
			int dx = max(w - 320, 0) >> 1;
			int dy = max(h - 240, 0) >> 1;
			surf->print(font_7x13, dx, dy, fps_str, 255, 0, true);
			surf->flip(dx, dy);
		}
	}
	else
	{
		surf->print(font_7x13, 0, 0, fps_str, 255, 0, true);
		surf->flip();
	}

	/*
	if (w > 320 || h > 240) surf->flip(x, y, w < 320 ? 320 : w, h < 240 ? 240 : h);
	else surf->flip(x, y);

	XBuf = surf->data();
	*/
}


int draw_scanline()
{
	int irq = 0;

	int dispmin = (MaxLine-MinLine>MAXDISP ? MinLine+((MaxLine-MinLine-MAXDISP+1)>>1) : MinLine);
	int dispmax = (MaxLine-MinLine>MAXDISP ? MaxLine-((MaxLine-MinLine-MAXDISP+1)>>1) : MaxLine);

	scanline=(scanline+1)%SCANLINES_PER_FRAME;
	io.vdc_status&=~VDC_RasHit;

	if (scanline>MaxLine) io.vdc_status|=VDC_InVBlank;

	if (scanline==MinLine)
	{
		io.vdc_status&=~VDC_InVBlank;
		prevline=dispmin;
		ScrollYDiff = 0;
		oldScrollYDiff = 0;
	}
	else if (scanline==MaxLine)
	{
		if (CheckSprites()) io.vdc_status|=VDC_SpHit;
		else io.vdc_status&=~VDC_SpHit;

		if (prevline<dispmax)
		{
			RefreshLine(prevline,dispmax);
			if (SpriteON && SPONSwitch) RefreshSprite(prevline,dispmax+1);
		}
		prevline=dispmax+1;

		sound_mix();
		RefreshScreen();
		//sound_mix();
	}

	if (scanline>=MinLine && scanline<=MaxLine)
	{
		if (scanline==(io.VDC[RCR].W&1023)-64)
		{
			if (RasHitON && dispmin<=scanline && scanline<=dispmax)
			{
				if (prevline<dispmax)
				{
					RefreshLine(prevline,scanline-1);
					if (SpriteON && SPONSwitch) RefreshSprite(prevline,scanline);
				}
				prevline=scanline;
			}
			io.vdc_status|=VDC_RasHit;
			if (RasHitON) irq = 1;
		}
		else if (scroll)
		{
			if (scanline-1>prevline)
			{
				int	tmpScrollX, tmpScrollY, tmpScrollYDiff;
				tmpScrollX=ScrollX;
				tmpScrollY=ScrollY;
				tmpScrollYDiff=ScrollYDiff;
				ScrollX=oldScrollX;
				ScrollY=oldScrollY;
				ScrollYDiff=oldScrollYDiff;
				RefreshLine(prevline,scanline-2);
				if (SpriteON && SPONSwitch) RefreshSprite(prevline,scanline-1);
				prevline=scanline-1;
				ScrollX = tmpScrollX;
				ScrollY = tmpScrollY;
				ScrollYDiff = tmpScrollYDiff;
			}
		}
	}
	else
	{
		int rcr = (io.VDC[RCR].W&1023)-64;
		if (scanline==rcr && RasHitON)
		{
			io.vdc_status |= VDC_RasHit;
			irq = 1;
		}
	}

	scroll=0;

	if (scanline==MaxLine+1)
	{
		{
			io.JOY[0] = 0;
			io.JOY[1] = 0;
			io.JOY[2] = 0;
			io.JOY[3] = 0;
			io.JOY[4] = 0;

			using namespace gp2x::joystick;

			int joy = read_state();
			if (joy < 0 || (joy & BUTTON_Y) != 0 ) return 0;

			if (joy & BUTTON_START) io.JOY[0] |= JOY_RUN;
 			if (joy & BUTTON_SELECT) io.JOY[0] |= JOY_SELECT;
 			if (joy & BUTTON_UP) io.JOY[0] |= JOY_UP;
 			if (joy & BUTTON_DOWN) io.JOY[0] |= JOY_DOWN;
 			if (joy & BUTTON_LEFT) io.JOY[0] |= JOY_LEFT;
 			if (joy & BUTTON_RIGHT) io.JOY[0] |= JOY_RIGHT;
 			if (joy & (BUTTON_A | BUTTON_X)) io.JOY[0] |= JOY_A;
 			if (joy & BUTTON_B) io.JOY[0] |= JOY_B;

			static bool accept_l = true;

 			if (joy & BUTTON_L)
 			{
 				if  (accept_l)
 				{
 					scale_to_fit = !scale_to_fit, accept_l = false;
 					surf->clear(true);
 				}
 			}
 			else
 				accept_l = true;


 			if ((joy & BUTTON_VOL_DOWN) && !(joy & BUTTON_VOL_UP) && sound_volume > 0)
 			{
 				--sound_volume;
 				gp2x::sound::volume(sound_volume, sound_volume);
 			}
 			if ((joy & BUTTON_VOL_UP) && !(joy & BUTTON_VOL_DOWN) && sound_volume < 100)
 			{
 				++sound_volume;
 				gp2x::sound::volume(sound_volume, sound_volume);
 			}
		}

		if (io.vdc_satb==1 || io.VDC[DCR].W&0x0010)
		{
			memcpy(SPRAM,VRAM+io.VDC[SATB].W*2,64*8);
			io.vdc_satb=1;
			io.vdc_status&=~VDC_SATBfinish;
		}

		if (irq) io.vdc_pendvsync = 1;
		else
		{
			if (VBlankON) irq = 1;
		}
	}
	else if (scanline==min(MaxLine+5, SCANLINES_PER_FRAME-1))
	{
		if (io.vdc_satb)
		{
			io.vdc_status|=VDC_SATBfinish;
			io.vdc_satb = 0;
			if (SATBIntON) irq = 1;
		}
	}
	else if (io.vdc_pendvsync && !irq)
	{
		io.vdc_pendvsync = 0;
		if (VBlankON) irq = 1;
	}

	if (irq && !(irq_mask&IRQ1)) irq_status|=IRQ1;

	return 1;
}

#define	VRAMSIZE	0x20000

void reset_video()
{
	memset(WRAM,0,0x2000);
	memset(vchange,1,VRAMSIZE/32);
	memset(vchanges,1,VRAMSIZE/128);

	XBuf = surf_pce->data();
	//XBuf = surf->data();
}

void init_video()
{
	WRAM=(byte *)malloc(0x2000);
	VRAM=(byte *)malloc(VRAMSIZE);
	VRAM2=(unsigned long *)malloc(VRAMSIZE);
	vchange = (byte *)malloc(VRAMSIZE/32);
	VRAMS=(unsigned long *)malloc(VRAMSIZE);
	vchanges = (byte *)malloc(VRAMSIZE/128);

	reset_video();
}

void trash_video()
{
	if (vchange) free(vchange);

	if (vchanges) free(vchanges);
	if (VRAMS) free(VRAMS);

	if (WRAM) free(WRAM);
	if (VRAM) free(VRAM);
	if (VRAM2) free(VRAM2);
}
